/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::ArmarXObjects::EmergencyStop
 * @author     Stefan Reither ( stef dot reither at web dot de )
 * @date       2016
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "EmergencyStopWidget.h"

#include <QGridLayout>
#include <QShortcut>
#include <QTimer>


#define EMERGENCY_STOP_PROXY "EmergencyStopMaster"
#define EMERGENCY_STOP_TOPIC_NAME "EmergencyStop"

namespace armarx
{
    EmergencyStopWidget::EmergencyStopWidget(QWidget* parent, ArmarXMainWindow* mainWindow) :
        mainWindow(mainWindow),
        iconNormal(QPixmap::fromImage(QImage(QString::fromUtf8(":/icons/emergency-stop.png")))),
        iconDark(QPixmap::fromImage(QImage(QString::fromUtf8(":/icons/emergency-stop-dark.png")))),
        timer(new QTimer(this))
    {
        qRegisterMetaType<EmergencyStopState>("EmergencyStopState");

        // Redundant timer apart from the topic in order to recover from bad states if a topic
        // message was lost.
        timer->setInterval(std::chrono::milliseconds(2'000));
        connect(timer, &QTimer::timeout, this, &EmergencyStopWidget::updateEmergencyStopState);
        connect(this,
                &EmergencyStopWidget::startPeriodicStateUpdate,
                timer,
                qOverload<>(&QTimer::start));
        connect(this, &EmergencyStopWidget::stopPeriodicStateUpdate, timer, &QTimer::stop);
        QIcon icon;
        icon.addPixmap(iconNormal, QIcon::Normal, QIcon::Off);
        icon.addPixmap(iconDark, QIcon::Normal, QIcon::On);

        button = new QToolButton();
        button->setCheckable(true);
        button->setIcon(icon);
        button->setIconSize(QSize(68, 28));
        button->setToolTip(QString::fromStdString("Controls the EmergencyStop. When pressed the "
                                                  "EmergencyStop is active. Shortcut: Pause Key"));
        button->setVisible(false);
        QGridLayout* l = new QGridLayout(this->getWidget());
        this->getWidget()->setLayout(l);
        this->getWidget()->layout()->addWidget(button);

        QShortcut* enableSS2Shortcut1 = new QShortcut(this->getWidget());
        enableSS2Shortcut1->setContext(Qt::ApplicationShortcut);
        enableSS2Shortcut1->setKey(Qt::Key_Pause);
        QShortcut* enableSS2Shortcut2 = new QShortcut(this->getWidget());
        enableSS2Shortcut2->setContext(Qt::ApplicationShortcut);
        enableSS2Shortcut2->setKey(Qt::Key_End);
        connect(enableSS2Shortcut1, &QShortcut::activated, this, &EmergencyStopWidget::enableSS2);
        connect(enableSS2Shortcut2, &QShortcut::activated, this, &EmergencyStopWidget::enableSS2);

        QShortcut* releaseSS2Shortcut1 = new QShortcut(this->getWidget());
        releaseSS2Shortcut1->setContext(Qt::ApplicationShortcut);
        releaseSS2Shortcut1->setKey(Qt::SHIFT | Qt::Key_Pause);
        QShortcut* releaseSS2Shortcut2 = new QShortcut(this->getWidget());
        releaseSS2Shortcut2->setContext(Qt::ApplicationShortcut);
        releaseSS2Shortcut2->setKey(Qt::SHIFT | Qt::Key_End);
        connect(releaseSS2Shortcut1, &QShortcut::activated, this, &EmergencyStopWidget::releaseSS2);
        connect(releaseSS2Shortcut2, &QShortcut::activated, this, &EmergencyStopWidget::releaseSS2);

        connect(button, &QPushButton::clicked, this, &EmergencyStopWidget::clicked);
        std::stringstream str;
        str << "After EmergencyStop was activated, you have to wait "
            << deactivationWaitPeriod.count() << " ms before you can deactivate it.";
        button->setToolTip(QString::fromStdString(str.str()));
    }

    QWidget*
    EmergencyStopWidget::getButtonWidget()
    {
        return this->getWidget();
    }

    void
    EmergencyStopWidget::onInitComponent()
    {
        usingProxy(EMERGENCY_STOP_PROXY);
        usingTopic(EMERGENCY_STOP_TOPIC_NAME);
    }

    void
    EmergencyStopWidget::onConnectComponent()
    {
        ARMARX_INFO << "SS2 widget connected.";
        emergencyStopMasterPrx = getProxy<EmergencyStopMasterInterfacePrx>(EMERGENCY_STOP_PROXY);
        QMetaObject::invokeMethod(button, "setVisible", Qt::QueuedConnection, Q_ARG(bool, true));
        EmergencyStopState state = EmergencyStopState::eEmergencyStopActive;

        try
        {
            state = emergencyStopMasterPrx->getEmergencyStopState();
        }
        catch (Ice::Exception const& e)
        {
            ARMARX_ERROR << "Could not query SS2 state." << deactivateSpam(2);
        }

        QMetaObject::invokeMethod(
            this, "setChecked", Qt::QueuedConnection, Q_ARG(EmergencyStopState, state));

        emit startPeriodicStateUpdate();
    }

    void
    EmergencyStopWidget::onDisconnectComponent()
    {
        ARMARX_IMPORTANT << "SS2 widget disconnected.  This is expected if the earlier connected "
                            "robot unit shut down.";
        QMetaObject::invokeMethod(button, "setVisible", Qt::QueuedConnection, Q_ARG(bool, false));

        emit stopPeriodicStateUpdate();
    }

    void
    EmergencyStopWidget::loadSettings(QSettings* settings)
    {
        ;
    }

    void
    EmergencyStopWidget::saveSettings(QSettings* settings)
    {
        ;
    }

    void
    EmergencyStopWidget::reportEmergencyStopState(EmergencyStopState state, const Ice::Current&)
    {
        QMetaObject::invokeMethod(
            this, "setChecked", Qt::QueuedConnection, Q_ARG(EmergencyStopState, state));
    }

    void
    EmergencyStopWidget::enableSS2()
    {
        if (emergencyStopMasterPrx)
        {
            emergencyStopMasterPrx->setEmergencyStopState(EmergencyStopState::eEmergencyStopActive);
        }
    }

    void
    EmergencyStopWidget::releaseSS2()
    {
        if (clock_t::now() > timeLastActivated + deactivationWaitPeriod)
        {
            if (emergencyStopMasterPrx)
            {
                emergencyStopMasterPrx->setEmergencyStopState(
                    EmergencyStopState::eEmergencyStopInactive);
            }
        }
        else
        {
            button->setChecked(true);
        }
    }

    void
    EmergencyStopWidget::clicked(bool checked)
    {
        if (emergencyStopMasterPrx)
        {
            EmergencyStopState const state = emergencyStopMasterPrx->getEmergencyStopState();
            bool success = true;
            switch (state)
            {
                case EmergencyStopState::eEmergencyStopActive:
                    if (clock_t::now() > timeLastActivated + deactivationWaitPeriod)
                    {
                        if (emergencyStopMasterPrx)
                        {
                            // Only release SS2 if the state we received still is "Active".  Fail
                            // otherwise.
                            bool const success =
                                emergencyStopMasterPrx->checkAndSetEmergencyStopState(
                                    EmergencyStopState::eEmergencyStopInactive, state);
                            if (not success)
                            {
                                ARMARX_WARNING << "Toggling SS2 failed since it probably has "
                                                  "already been activated from another session.  "
                                                  "This mechanism is to prevent two simultanous "
                                                  "toggles of the SS2 to cancel out.";
                            }
                        }
                    }
                    else
                    {
                        button->setChecked(true);
                    }
                    break;
                case EmergencyStopState::eEmergencyStopInactive:
                    if (emergencyStopMasterPrx)
                    {
                        // Always enable SS2 without checking.
                        enableSS2();
                    }
                    break;
            }
            if (not success)
            {
                ARMARX_WARNING << "Releasing SS2 failed.";
            }
        }
    }

    std::string
    EmergencyStopWidget::getDefaultName() const
    {
        return "EmergencyStopGuiWidget" + iceNameUUID;
    }

    void
    EmergencyStopWidget::setChecked(const EmergencyStopState state)
    {
        ARMARX_INFO << "STATE IS :"
                    << (state == EmergencyStopState::eEmergencyStopActive ? "active" : "inactive");
        switch (state)
        {
            case EmergencyStopState::eEmergencyStopActive:
                button->setChecked(true);
                break;
            case EmergencyStopState::eEmergencyStopInactive:
                button->setChecked(false);
                timeLastActivated = clock_t::now();
                break;
            default:
                button->setChecked(false);
        }
    }

    void
    EmergencyStopWidget::updateEmergencyStopState()
    {
        ARMARX_INFO << "UPDATE!";
        try
        {
            if (emergencyStopMasterPrx)
            {
                setChecked(emergencyStopMasterPrx->getEmergencyStopState());
            }
        }
        catch (Ice::Exception const& e)
        {
            ARMARX_ERROR << "Could not query SS2 state." << deactivateSpam(2);
            setChecked(EmergencyStopState::eEmergencyStopActive);
        }
    }

} // namespace armarx
